import sys
import socket

from requests.adapters import HTTPAdapter
from urllib3 import poolmanager

KEEP_INTERVAL = 30
KEEP_COUNT = 3
KEEP_IDLE = 30
SOCKET_TCP = socket.IPPROTO_TCP

core_socket_options = [
    (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
    (SOCKET_TCP, socket.TCP_NODELAY, 1)
]


class KeepAliveAdapter(HTTPAdapter):
    """
    Extended HTTP adapter that sets preferred keep alive options
    """

    # pylint: disable=no-member
    def __init__(self, **kwargs):
        self.socket_options = core_socket_options.copy()
        interval = kwargs.pop('keep_interval', KEEP_INTERVAL)
        count = kwargs.pop('keep_count', KEEP_COUNT)
        idle = kwargs.pop('keep_idle', KEEP_IDLE)

        if getattr(socket, 'TCP_KEEPINTVL', None) is not None:
            self.socket_options.append((SOCKET_TCP, socket.TCP_KEEPINTVL, interval))
        if getattr(socket, 'TCP_KEEPCNT', None) is not None:
            self.socket_options.append((SOCKET_TCP, socket.TCP_KEEPCNT, count))
        if getattr(socket, 'TCP_KEEPIDLE', None) is not None:
            self.socket_options.append((SOCKET_TCP, socket.TCP_KEEPIDLE, idle))
        if sys.platform == 'darwin':
            self.socket_options.append((SOCKET_TCP, getattr(socket, 'TCP_KEEPALIVE', 0x10), interval))
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        self.poolmanager = poolmanager.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            socket_options=self.socket_options,
            **pool_kwargs)



from typing import Any, Sequence

from clickhouse_connect.datatypes import registry
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.common import read_leb128, read_leb128_str, write_leb128
from clickhouse_connect.driver.query import DataResult
from clickhouse_connect.driver.transform import DataTransform, QueryContext


class NativeTransform(DataTransform):
    # pylint: disable=too-many-locals
    def _transform_response(self, source: Sequence, context: QueryContext) -> DataResult:
        if not isinstance(source, memoryview):
            source = memoryview(source)
        loc = 0
        names = []
        col_types = []
        result = []
        total_size = len(source)
        block = 0
        use_none = context.use_none
        while loc < total_size:
            result_block = []
            num_cols, loc = read_leb128(source, loc)
            num_rows, loc = read_leb128(source, loc)
            for col_num in range(num_cols):
                name, loc = read_leb128_str(source, loc)
                if block == 0:
                    names.append(name)
                type_name, loc = read_leb128_str(source, loc)
                if block == 0:
                    col_type = registry.get_from_name(type_name)
                    col_types.append(col_type)
                else:
                    col_type = col_types[col_num]
                context.start_column(name, col_type)
                column, loc = col_type.read_native_column(source, loc, num_rows, use_none=use_none)
                result_block.append(column)
            block += 1
            result.extend(list(zip(*result_block)))
        return DataResult(result, tuple(names), tuple(col_types))

    def build_insert(self, data: Sequence[Sequence[Any]], *, column_names: Sequence[str],
                     column_type_names: Sequence[str] = None,
                     column_types: Sequence[ClickHouseType] = None,
                     column_oriented: bool = False):
        if not column_types:
            column_types = [registry.get_from_name(name) for name in column_type_names]
        output = bytearray()
        columns = data if column_oriented else tuple(zip(*data))
        write_leb128(len(columns), output)
        write_leb128(len(columns[0]), output)
        for col_name, col_type, column in zip(column_names, column_types, columns):
            write_leb128(len(col_name), output)
            output += col_name.encode()
            write_leb128(len(col_type.name), output)
            output += col_type.name.encode()
            col_type.write_native_column(column, output)
        return output



import logging
import pytz

from abc import ABCMeta, abstractmethod
from typing import Iterable, Tuple, Optional, Any, Union, Sequence, Dict
from pytz.exceptions import UnknownTimeZoneError

from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.exceptions import ProgrammingError, InternalError
from clickhouse_connect.driver.models import ColumnDef, SettingDef
from clickhouse_connect.driver.query import QueryResult, np_result, to_pandas_df, from_pandas_df, to_arrow, \
    QueryContext, arrow_buffer

logger = logging.getLogger(__name__)


class Client(metaclass=ABCMeta):
    """
    Base ClickHouse Connect client
    """
    column_inserts = False
    valid_transport_settings = set()

    def __init__(self, database: str, query_limit: int, uri: str):
        """
        Shared initialization of ClickHouse Connect client
        :param database: database name
        :param query_limit: default LIMIT for queries
        :param uri: uri for error messages
        """
        self.limit = query_limit
        self.server_tz = pytz.UTC
        self.server_version, server_tz, self.database = \
            tuple(self.command('SELECT version(), timezone(), database()', use_database=False))
        try:
            self.server_tz = pytz.timezone(server_tz)
        except UnknownTimeZoneError:
            logger.warning('Warning, server is using an unrecognized timezone %s, will use UTC default', server_tz)
        server_settings = self.query('SELECT name, value, changed, description, type, readonly FROM system.settings')
        self.server_settings = {row['name']: SettingDef(**row) for row in server_settings.named_results()}
        if database and not database == '__default__':
            self.database = database
        self.uri = uri

    def _validate_settings(self, settings: Optional[Dict[str, Any]], stringify: bool = False) -> Dict[str, Any]:
        """
        This strips any ClickHouse settings that are not recognized or are read only
        :param settings:  Dictionary of setting name and values
        :param stringify:  Return the result dictionary values as strings
        :return:
        """
        validated = {}
        if settings:
            for key, value in settings.items():
                if key not in self.valid_transport_settings:
                    setting_def = self.server_settings.get(key)
                    if setting_def is None or setting_def.readonly:
                        logger.debug('Setting %s is not valid or read only, ignoring', key)
                        continue
                if stringify:
                    if isinstance(value, bool):
                        value = '1' if value else '0'
                    else:
                        value = str(value)
                validated[key] = value
        return validated

    def _prep_query(self, context: QueryContext):
        if context.is_select and not context.has_limit and self.limit:
            return f'{context.final_query}\n LIMIT {self.limit}'
        return context.final_query

    @abstractmethod
    def _query_with_context(self, context: QueryContext):
        pass

    @abstractmethod
    def client_setting(self, name, value):
        """
        Set a clickhouse setting for the client after initialization
        :param name: Setting name
        :param value: Setting value
        """

    # pylint: disable=duplicate-code,too-many-arguments
    def query(self,
              query: str = None,
              parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
              settings: Optional[Dict[str, Any]] = None,
              query_formats: Optional[Dict[str, str]] = None,
              column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
              encoding: Optional[str] = None,
              use_none: bool = True,
              context: QueryContext = None) -> QueryResult:
        """
        Main query method for SELECT, DESCRIBE and other commands that result a result matrix
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param query_formats: See QueryContext __init__ docstring
        :param column_formats: See QueryContext __init__ docstring
        :param encoding: See QueryContext __init__ docstring
        :param use_none: Use None for ClickHouse nulls instead of empty values
        :param context An alternative QueryContext parameter object that contains some or all of the method arguments
        :return: QueryResult -- data and metadata from response
        """
        if context:
            query_context = context.updated_copy(query,
                                                 parameters,
                                                 settings,
                                                 query_formats,
                                                 column_formats,
                                                 encoding,
                                                 self.server_tz,
                                                 False)
        else:
            query_context = QueryContext(query,
                                         parameters,
                                         settings,
                                         query_formats,
                                         column_formats,
                                         encoding,
                                         self.server_tz,
                                         use_none)
        return self._query_with_context(query_context)

    @abstractmethod
    def raw_query(self,
                  query: str,
                  parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  fmt: str = None) -> bytes:
        """
        Query method that simply returns the raw ClickHouse format bytes
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param fmt: ClickHouse output format
        :return: bytes representing raw ClickHouse return value based on format
        """

    # pylint: disable=duplicate-code
    def query_np(self,
                 query: str = None,
                 parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, str]] = None,
                 encoding: Optional[str] = None,
                 context: QueryContext = None):
        """
        Query method that returns the results as a numpy array
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param query_formats: See QueryContext __init__ docstring
        :param column_formats: See QueryContext __init__ docstring.
        :param encoding: See QueryContext __init__ docstring
        :param context An alternative QueryContext parameter object that contains some or all of the method arguments
        :return: Numpy array representing the result set
        """
        return np_result(self.query(query,
                                    parameters,
                                    settings,
                                    query_formats,
                                    column_formats,
                                    encoding,
                                    False,
                                    context))

    # pylint: disable=duplicate-code
    def query_df(self,
                 query: str = None,
                 parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, str]] = None,
                 encoding: Optional[str] = None,
                 context: QueryContext = None):
        """
        Query method that results the results as a pandas dataframe
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param query_formats: See QueryContext __init__ docstring
        :param column_formats: See QueryContext __init__ docstring
        :param encoding: See QueryContext __init__ docstring
        :param context An alternative QueryContext parameter object that contains some or all of the method arguments
        :return: Numpy array representing the result set
        """
        return to_pandas_df(self.query(query,
                                       parameters,
                                       settings,
                                       query_formats,
                                       column_formats,
                                       encoding,
                                       False,
                                       context))

    def query_arrow(self,
                    query: str,
                    parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                    settings: Optional[Dict[str, Any]] = None,
                    use_strings: bool = True):
        """
        Query method using the ClickHouse Arrow format to return a PyArrow table
        :param query: Query statement/format string
        :param parameters: Optional dictionary used to format the query
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_strings:  Convert ClickHouse String type to Arrow string type (instead of binary)
        :return: PyArrow.Table
        """
        arrow_settings = {} if not settings else settings.copy()
        if 'output_format_arrow_string_as_string' not in arrow_settings:
            arrow_settings['output_format_arrow_string_as_string'] = '1' if use_strings else '0'
        return to_arrow(self.raw_query(query, parameters, arrow_settings, 'Arrow'))

    @abstractmethod
    def command(self,
                cmd: str,
                parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                data: Union[str, bytes] = None,
                settings: Dict[str, Any] = None,
                use_database: bool = True) -> Union[str, int, Sequence[str]]:
        """
        Client method that returns a single value instead of a result set
        :param cmd: ClickHouse query/command as a python format string
        :param parameters: Optional dictionary of key/values pairs to be formatted
        :param data: Optional 'data' for the command (for INSERT INTO in particular)
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :param use_database: Send the database parameter to ClickHouse so the command will be executed in that database
         context.  Otherwise, no database will be specified with the command.  This is useful for determining
         the default user database
        :return: Decoded response from ClickHouse as either a string, int, or sequence of strings
        """

    @abstractmethod
    def ping(self) -> bool:
        """
        Validate the connection, does not throw an Exception (see debug logs)
        :return: ClickHouse server is up and reachable
        """

    # pylint: disable=too-many-arguments
    def insert(self,
               table: str,
               data: Sequence[Sequence[Any]],
               column_names: Union[str, Iterable[str]] = '*',
               database: str = '',
               column_types: Sequence[ClickHouseType] = None,
               column_type_names: Sequence[str] = None,
               column_oriented: bool = False,
               settings: Optional[Dict[str, Any]] = None):
        """
        Method to insert multiple rows/data matrix of native Python objects
        :param table: Target table
        :param data: Sequence of sequences of Python data
        :param column_names: Ordered list of column names or '*' if column types should be retrieved from ClickHouse table definition
        :param database: Target database -- will use client default database if not specified
        :param column_types: ClickHouse column types.  If set then column data does not need to be retrieved from the server
        :param column_type_names: ClickHouse column type names.  If set then column data does not need to be retrieved from the server
        :param column_oriented: If true the data is already "pivoted" in column form
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :return: No return, throws an exception if the insert fails
        """
        if len(data) == 0 or len(data[0]) == 0:
            logger.debug('Attempted insert with no data')
            return
        table, database, full_table = self.normalize_table(table, database)
        if isinstance(column_names, str):
            if column_names == '*':
                column_defs = [cd for cd in self.table_columns(table, database)
                               if cd.default_type not in ('ALIAS', 'MATERIALIZED')]
                column_names = [cd.name for cd in column_defs]
                column_types = [cd.ch_type for cd in column_defs]
            else:
                column_names = [column_names]
        elif len(column_names) == 0:
            raise ValueError('Column names must be specified for insert')
        if column_types is None:
            if column_type_names:
                column_types = [get_from_name(name) for name in column_type_names]
            else:
                column_map: Dict[str: ColumnDef] = {d.name: d for d in self.table_columns(table, database)}
                try:
                    column_types = [column_map[name].ch_type for name in column_names]
                except KeyError as ex:
                    raise ProgrammingError(f'Unrecognized column {ex} in table {table}') from None
        assert len(column_names) == len(column_types)
        self.data_insert(full_table, column_names, data, column_types, settings, column_oriented)

    def insert_df(self, table: str, data_frame, database: str = None, settings: Optional[Dict] = None):
        """
        Insert a pandas DataFrame into ClickHouse
        :param table: ClickHouse table
        :param data_frame: two-dimensional pandas dataframe
        :param database: Optional ClickHouse database
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :return: No return, throws an exception if the insert fails
        """
        return self.insert(table, database=database, settings=settings, **from_pandas_df(data_frame))

    def insert_arrow(self,
                     table: str,
                     arrow_table,
                     database: str = None,
                     settings: Optional[Dict] = None):
        """
        Insert a PyArrow table DataFrame into ClickHouse using raw Arrow format
        :param table: ClickHouse table
        :param arrow_table: PyArrow Table object
        :param database: Optional ClickHouse database
        :param settings: Optional dictionary of ClickHouse settings (key/string values)
        :return: No return, throws an exception if the insert fails
        """
        _, _, full_table = self.normalize_table(table, database)
        column_names, insert_block = arrow_buffer(arrow_table)
        self.raw_insert(full_table, column_names, insert_block, settings, 'Arrow')

    def normalize_table(self, table: str, database: Optional[str]) -> Tuple[str, str, str]:
        """
        Convenience method to return the table, database, and full table name
        :param table: table name
        :param database: optional database
        :return: Tuple of bare table name, bare database name, and full database.table
        """
        split = table.split('.')
        if len(split) > 1:
            full_name = table
            database = split[0]
            table = split[1]
        else:
            name = table
            database = database or self.database
            full_name = f'{database}.{name}'
        return table, database, full_name

    def min_version(self, version_str: str) -> bool:
        """
        Determine whether the connected server is at least the submitted version
        :param version_str:  Version string consisting of up to 4 integers delimited by dots
        :return:  True version_str is greater than the server_version, False if less than
        """
        try:
            server_parts = [int(x) for x in self.server_version.split('.')]
            server_parts.extend([0] * (4 - len(server_parts)))
            version_parts = [int(x) for x in version_str.split('.')]
            version_parts.extend([0] * (4 - len(version_parts)))
        except ValueError:
            logger.warning('Server %s or requested version %s does not match format of numbers separated by dots',
                           self.server_version, version_str)
            return False
        for x, y in zip(server_parts, version_parts):
            if x > y:
                return True
            if x < y:
                return False
        return True

    def table_columns(self, table: str, database: str) -> Tuple[ColumnDef]:
        """
        Return complete column definitions for a ClickHouse table
        :param table: table name
        :param database: database name
        :return: list of ColumnDef named tuples
        """
        column_result = self.query(f'DESCRIBE TABLE {database}.{table}')
        if not column_result.result_set:
            raise InternalError(f'No table columns found for {database}.{table}')
        return tuple(ColumnDef(**row) for row in column_result.named_results())

    @abstractmethod
    def data_insert(self,
                    table: str,
                    column_names: Iterable[str],
                    data: Iterable[Iterable[Any]],
                    column_types: Iterable[ClickHouseType],
                    settings: Optional[Dict] = None,
                    column_oriented: bool = False):
        """
        Subclass implementation of the data insert
        :param table: ClickHouse table
        :param column_names: List of ClickHouse columns
        :param data: Data matrix
        :param column_types: Parallel list of ClickHouseTypes to insert
        :param settings:  Optional dictionary of ClickHouse settings (key/string values)
        :param column_oriented: Whether the data is already pivoted as a sequence of columns
        :return: No return, throws an exception if the insert fails
        """

    @abstractmethod
    def raw_insert(self, table: str,
                   column_names: Sequence[str],
                   insert_block: Union[str,  bytes],
                   settings: Optional[Dict] = None,
                   fmt: Optional[str] = None):
        """
        Insert data already formatted in a bytes object
        :param table: Table name (whether qualified with the database name or not)
        :param column_names: Sequence of column names
        :param insert_block: Binary or string data already in a recognized ClickHouse format
        :param settings:  Optional dictionary of ClickHouse settings (key/string values)
        :param fmt: Valid clickhouse format
        """

    def close(self):
        """
        Subclass implementation to close the connection to the server/deallocate the client
        """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



from sqlalchemy.engine.default import DefaultDialect

from clickhouse_connect import dbapi
from clickhouse_connect.cc_sqlalchemy.sql import full_table
from clickhouse_connect.cc_sqlalchemy.sql.ddlcompiler import ChDDLCompiler
from clickhouse_connect.cc_sqlalchemy import ischema_names, reflect, dialect_name
from clickhouse_connect.cc_sqlalchemy.sql.preparer import ChIdentifierPreparer


# pylint: disable-msg=too-many-public-methods
class ClickHouseDialect(DefaultDialect):
    """
    See :py:class:`sqlalchemy.engine.interfaces`
    """
    name = dialect_name
    driver = 'connect'

    default_schema_name = 'default'
    supports_native_decimal = True
    supports_native_boolean = True
    returns_unicode_strings = True
    postfetch_lastrowid = False
    ddl_compiler = ChDDLCompiler
    preparer = ChIdentifierPreparer
    description_encoding = None
    max_identifier_length = 127
    ischema_names = ischema_names

    # pylint: disable=method-hidden
    @classmethod
    def dbapi(cls):
        return dbapi

    def initialize(self, connection):
        pass

    @staticmethod
    def get_schema_names(connection, **_):
        return [row.name for row in connection.execute('SHOW DATABASES')]

    @staticmethod
    def has_database(connection, db_name):
        return (connection.execute(f"SELECT name FROM system.databases WHERE name = '{db_name}'")).rowcount > 0

    def get_table_names(self, connection, schema=None, **kw):
        cmd = 'SHOW TABLES'
        if schema:
            cmd += ' FROM ' + schema
        return [row.name for row in connection.execute(cmd)]

    get_columns = staticmethod(reflect.get_columns)
    reflecttable = staticmethod(reflect.reflect_table)

    def get_primary_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_pk_constraint(self, conn, table_name, schema=None, **kw):
        return []

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_temp_table_names(self, connection, schema=None, **kw):
        return []

    def get_view_names(self, connection, schema=None, **kw):
        return []

    def get_temp_view_names(self, connection, schema=None, **kw):
        return []

    def get_view_definition(self, connection, view_name, schema=None, **kw):
        pass

    def get_indexes(self, connection, table_name, schema=None, **kw):
        return []

    def get_unique_constraints(self, connection, table_name, schema=None, **kw):
        return []

    def get_check_constraints(self, connection, table_name, schema=None, **kw):
        return []

    def has_table(self, connection, table_name, schema=None):
        result = connection.execute(f'EXISTS TABLE {full_table(table_name, schema)}')
        row = result.fetchone()
        return row[0] == 1

    def has_sequence(self, connection, sequence_name, schema=None):
        return False

    def do_begin_twophase(self, connection, xid):
        raise NotImplementedError

    def do_prepare_twophase(self, connection, xid):
        raise NotImplementedError

    def do_rollback_twophase(self, connection, xid, is_prepared=True, recover=False):
        raise NotImplementedError

    def do_commit_twophase(self, connection, xid, is_prepared=True, recover=False):
        raise NotImplementedError

    def do_recover_twophase(self, connection):
        raise NotImplementedError

    def set_isolation_level(self, dbapi_conn, level):
        pass

    def get_isolation_level(self, dbapi_conn):
        return None



import re

from superset.utils.core import GenericDataType
from clickhouse_connect.cc_sqlalchemy.datatypes.base import sqla_type_map
from clickhouse_connect.datatypes.format import set_default_formats

type_mapping = (
    (r'^(FLOAT|DECIMAL|INT|UINT)', GenericDataType.NUMERIC),
    (r'^DATE', GenericDataType.TEMPORAL),
    (r'^BOOL', GenericDataType.BOOLEAN)
)


def configure_types():
    """
    Monkey patch the Superset generic_type onto the clickhouse type, also set defaults for certain type formatting to be
    better compatible with superset
    """
    set_default_formats(FixedString='string',
                        IPv4='string',
                        UInt64='signed',
                        UUID='string')
    compiled = [(re.compile(pattern, re.IGNORECASE), gen_type) for pattern, gen_type in type_mapping]
    for name, sqla_type in sqla_type_map.items():
        for pattern, gen_type in compiled:
            match = pattern.match(name)
            if match:
                sqla_type.generic_type = gen_type
                break
        else:
            sqla_type.generic_type = GenericDataType.STRING



import sqlalchemy.schema as sa_schema
from sqlalchemy.orm.exc import NoResultFound

from clickhouse_connect.cc_sqlalchemy.datatypes.base import sqla_type_from_name
from clickhouse_connect.cc_sqlalchemy import dialect_name as dn
from clickhouse_connect.cc_sqlalchemy.ddl.tableengine import build_engine
from clickhouse_connect.cc_sqlalchemy.sql import full_table

ch_col_args = ('default_type', 'codec_expression', 'ttl_expression')


def get_columns(connection, table_name, schema=None, **_kwargs):
    """
    Build SQLAlchemy column dictionary (see
    :param connection: SQLAlchemy connection
    :param table_name: ClickHouse table
    :param schema: ClickHouse database name
    :param _kwargs: Ignored
    :return:
    """
    table_id = full_table(table_name, schema)
    result_set = connection.execute(f'DESCRIBE TABLE {table_id}')
    if not result_set:
        raise NoResultFound(f'Table {full_table} does not exist')
    columns = []
    for row in result_set:
        sqla_type = sqla_type_from_name(row.type)
        col = {'name': row.name,
               'type': sqla_type,
               'nullable': sqla_type.nullable,
               'autoincrement': False,
               'default': row.default_expression,
               'default_type': row.default_type,
               'comment': row.comment,
               'codec_expression': row.codec_expression,
               'ttl_expression': row.ttl_expression}
        columns.append(col)
    return columns


def get_engine(connection, table_name, schema=None):
    result_set = connection.execute(
        f"SELECT engine_full FROM system.tables WHERE database = '{schema}' and name = '{table_name}'")
    row = next(result_set, None)
    if not row:
        raise NoResultFound(f'Table {schema}.{table_name} does not exist')
    return build_engine(row.engine_full)


def reflect_table(connection, table, include_columns, exclude_columns, *_args, **_kwargs):
    schema = table.schema
    for col in get_columns(connection, table.name, schema):
        name = col.pop('name')
        if (include_columns and name not in include_columns) or (exclude_columns and name in exclude_columns):
            continue
        col_type = col.pop('type')
        col_args = {f'{dn}_{key}' if key in ch_col_args else key: value for key, value in col.items() if value}
        table.append_column(sa_schema.Column(name, col_type, **col_args))
    table.engine = get_engine(connection, table.name, schema)



from abc import ABC, abstractmethod
from typing import Sequence, Any

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.query import DataResult, QueryContext


_EMPTY_CONTEXT = QueryContext()


class DataTransform(ABC):

    def parse_response(self, source: Sequence, context: QueryContext = _EMPTY_CONTEXT) -> DataResult:
        """
        Decodes the ClickHouse byte buffer response into rows of native Python data
        :param source: A byte buffer or similar source
        :param context: The QueryContext to use in processing the response
        :return: DataResult -- data matrix, column names, column types
        """
        with context:
            return self._transform_response(source, context)

    @abstractmethod
    def build_insert(self, data: Sequence[Sequence[Any]], *, column_names: Sequence[str],
                     column_type_names: Sequence[str] = None,
                     column_types: Sequence[ClickHouseType] = None,
                     column_oriented: bool = False):
        """
        Encodes a dataset of Python sequences into a ClickHouse format
        :param data: Matrix of rows and columns of data
        :param column_names: Column names of the data to insert
        :param column_type_names: Column type names of the data
        :param column_types: Column types used to encode data in ClickHouse native format
        :param column_oriented: If true the dataset does not need to be "pivoted"
        :return: bytearray containing the dataset in the appropriate format
        """

    @abstractmethod
    def _transform_response(self, source: Sequence, context: QueryContext) -> DataResult:
        pass



from typing import Sequence, MutableSequence, Union

from clickhouse_connect.datatypes.base import ClickHouseType, TypeDef
from clickhouse_connect.driver.common import read_leb128, to_leb128


class String(ClickHouseType):
    python_null = ''

    def _from_row_binary(self, source, loc):
        length, loc = read_leb128(source, loc)
        return str(source[loc:loc + length], self.encoding), loc + length

    def _to_row_binary(self, value: str, dest: bytearray):
        value = bytes(value, self.encoding)
        dest += to_leb128(len(value)) + value

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        return self._read_native_impl(source, loc, num_rows, self.encoding)

    @staticmethod
    def _read_native_python(source, loc, num_rows, encoding: str):
        column = []
        app = column.append
        for _ in range(num_rows):
            length = 0
            shift = 0
            while True:
                b = source[loc]
                length += ((b & 0x7f) << shift)
                loc += 1
                if (b & 0x80) == 0:
                    break
                shift += 7
            app(str(source[loc: loc + length], encoding))
            loc += length
        return column, loc

    # pylint: disable=duplicate-code
    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        encoding = self.encoding
        app = dest.append
        if self.nullable:
            for x in column:
                if x is None:
                    app(0)
                else:
                    y = x.encode(encoding)
                    sz = len(y)
                    while True:
                        b = sz & 0x7f
                        sz >>= 7
                        if sz == 0:
                            app(b)
                            break
                        app(0x80 | b)
                    dest += y
        else:
            for x in column:
                y = x.encode(encoding)
                sz = len(y)
                while True:
                    b = sz & 0x7f
                    sz >>= 7
                    if sz == 0:
                        app(b)
                        break
                    app(0x80 | b)
                dest += y

    _read_native_impl = _read_native_python


class FixedString(ClickHouseType):
    valid_formats = 'string', 'native'

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._byte_size = type_def.values[0]
        self._name_suffix = type_def.arg_str
        self._empty_bytes = bytes(b'\x00' * self._byte_size)
        self.to_row_binary = self._to_rb_internal

    @property
    def python_null(self):
        return self._empty_bytes if self.read_format() == 'native' else ''

    def _from_row_binary(self, source: Sequence, loc: int):
        return bytes(source[loc:loc + self._byte_size]), loc + self._byte_size

    def _to_row_binary(self, value: Sequence, dest: MutableSequence):
        dest += value

    @property
    def _to_rb_internal(self):
        return self._to_row_binary_str if self.write_format() == 'string' else self._to_row_binary

    def _to_row_binary_str(self, value, dest: bytearray):
        value = str.encode(value, self.encoding)
        dest += value
        if len(value) < self._byte_size:
            dest += bytes((0,) * (self._byte_size - len(value)))

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        if self.read_format() == 'string':
            return self._read_native_str(source, loc, num_rows, self._byte_size, self.encoding)
        return self._read_native_bytes(source, loc, num_rows, self._byte_size)

    @staticmethod
    def _read_native_str_python(source: Sequence, loc: int, num_rows: int, sz: int, encoding: str):
        column = []
        app = column.append
        end = loc + sz * num_rows
        for ix in range(loc, end, sz):
            try:
                app(str(source[ix: ix + sz], encoding).rstrip('\x00'))
            except UnicodeDecodeError:
                app(source[ix: ix + sz].hex())
        return column, end

    @staticmethod
    def _read_native_bytes_python(source: Sequence, loc: int, num_rows: int, sz: int):
        end = loc + sz * num_rows
        return [bytes(source[ix: ix + sz]) for ix in range(loc, end, sz)], end

    # pylint: disable=too-many-branches
    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        ext = dest.extend
        sz = self._byte_size
        empty = bytes((0,) * sz)
        str_enc = str.encode
        enc = self.encoding
        first = self._first_value(column)
        if isinstance(first, str) or self.write_format() == 'string':
            if self.nullable:
                for x in column:
                    if x is None:
                        ext(empty)
                    else:
                        try:
                            b = str_enc(x, enc)
                        except UnicodeEncodeError:
                            b = empty
                        ext(b)
                        if len(b) < sz:
                            ext(empty[:-len(b)])
            else:
                for x in column:
                    try:
                        b = str_enc(x, enc)
                    except UnicodeEncodeError:
                        b = empty
                    ext(b)
                    if len(b) < sz:
                        ext(empty[:-len(b)])
        elif self.nullable:
            for x in column:
                if not x:
                    ext(empty)
                else:
                    ext(x)
        else:
            for x in column:
                ext(x)

    _read_native_str = _read_native_str_python
    _read_native_bytes = _read_native_bytes_python



import logging

import clickhouse_connect.datatypes.container
import clickhouse_connect.datatypes.network as dt_network
import clickhouse_connect.datatypes.numeric as dt_numeric
import clickhouse_connect.datatypes.special as dt_special
import clickhouse_connect.datatypes.string as dt_string
import clickhouse_connect.datatypes.temporal
import clickhouse_connect.datatypes.registry


logger = logging.getLogger(__name__)

# pylint: disable=protected-access
try:
    from clickhouse_connect.driverc import creaders

    dt_string.String._read_native_impl = creaders.read_string_column
    dt_string.FixedString._read_native_str = creaders.read_fixed_string_str
    dt_string.FixedString._read_native_bytes = creaders.read_fixed_string_bytes
except ImportError:
    logger.warning('Unable to connect optimized C driver functions, falling back to pure Python', exc_info=True)



import ipaddress
import re
import uuid
import pytz

from enum import Enum
from typing import NamedTuple, Any, Tuple, Dict, Sequence, Optional, Union
from datetime import date, datetime, tzinfo

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.datatypes.container import Array
from clickhouse_connect.datatypes.format import format_map
from clickhouse_connect.driver.options import HAS_NUMPY, HAS_PANDAS, check_pandas, check_numpy, HAS_ARROW, check_arrow
from clickhouse_connect.driver.threads import query_settings

if HAS_PANDAS:
    import pandas as pa

if HAS_NUMPY:
    import numpy as np

if HAS_ARROW:
    import pyarrow

limit_re = re.compile(r'\s+LIMIT($|\s)', re.IGNORECASE)
select_re = re.compile(r'(^|\s)SELECT\s', re.IGNORECASE)
insert_re = re.compile(r'(^|\s)INSERT\s*INTO', re.IGNORECASE)


# pylint: disable=too-many-instance-attributes
class QueryContext:
    """
    Argument/parameter object for queries.  This context is used to set thread/query specific formats
    """

    # pylint: disable=duplicate-code,too-many-arguments
    def __init__(self,
                 query: str = '',
                 parameters: Optional[Dict[str, Any]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                 encoding: Optional[str] = None,
                 server_tz: tzinfo = pytz.UTC,
                 use_none: bool = True):
        """
        Initializes various configuration settings for the query context

        :param query:  Query string with Python style format value replacements
        :param parameters: Optional dictionary of substitution values
        :param settings: Optional ClickHouse settings for the query
        :param query_formats: Optional dictionary of query formats with the key of a ClickHouse type name
          (with * wildcards) and a value of valid query formats for those types.
          The value 'encoding' can be sent to change the expected encoding for this query, with a value of
          the desired encoding such as `latin-1`
        :param column_formats: Optional dictionary of column specific formats.  The key is the column name,
          The value is either the format for the data column (such as 'string' for a UUID column) or a
          second level "format" dictionary of a ClickHouse type name and a value of query formats.  This
          secondary dictionary can be used for nested column types such as Tuples or Maps
        :param encoding: Optional string encoding for this query, such as 'latin-1'
        :param column_formats: Optional dictionary
        :param use_none:
        """
        self.query = query
        self.parameters = parameters or {}
        self.settings = settings or {}
        self.query_formats = query_formats or {}
        self.column_formats = column_formats or {}
        self.encoding = encoding
        self.server_tz = server_tz
        self.use_none = use_none
        self.final_query = finalize_query(query, parameters, server_tz)
        self._uncommented_query = None

    @property
    def uncommented_query(self) -> str:
        if not self._uncommented_query:
            self._uncommented_query = remove_sql_comments(self.final_query)
        return self._uncommented_query

    @property
    def is_select(self) -> bool:
        return select_re.search(self.uncommented_query) is not None

    @property
    def has_limit(self) -> bool:
        return limit_re.search(self.uncommented_query) is not None

    @property
    def is_insert(self) -> bool:
        return insert_re.search(self.uncommented_query) is not None

    def updated_copy(self,
                     query: Optional[str] = None,
                     parameters: Optional[Dict[str, Any]] = None,
                     settings: Optional[Dict[str, Any]] = None,
                     query_formats: Optional[Dict[str, str]] = None,
                     column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                     encoding: Optional[str] = None,
                     server_tz: Optional[tzinfo] = None,
                     use_none: Optional[bool] = None) -> 'QueryContext':
        """
        Creates Query context copy with parameters overridden/updated as appropriate
        """
        return QueryContext(query or self.query,
                            self.parameters.update(parameters or {}),
                            self.settings.update(settings or {}),
                            self.query_formats.update(query_formats or {}),
                            self.column_formats.update(column_formats or {}),
                            encoding if encoding else self.encoding,
                            server_tz if server_tz else self.server_tz,
                            use_none if use_none is not None else self.use_none)

    def __enter__(self):
        query_settings.query_overrides = format_map(self.query_formats)
        query_settings.query_encoding = self.encoding
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        query_settings.query_overrides = None
        query_settings.column_overrides = None
        query_settings.query_encoding = None

    def start_column(self, name: str, ch_type: ClickHouseType):
        if name in self.column_formats:
            fmts = self.column_formats[name]
            if isinstance(fmts, str):
                if isinstance(ch_type, Array):
                    fmt_map = {ch_type.element_type.__class__: fmts}
                else:
                    fmt_map = {ch_type.__class__: fmts}
            else:
                fmt_map = format_map(fmts)
            query_settings.column_overrides = fmt_map
        else:
            query_settings.column_overrides = None


class QueryResult:
    """
    Wrapper class for query return values and metadata
    """

    def __init__(self, result_set: Sequence[Sequence[Any]], column_names: Tuple[str, ...],
                 column_types: Tuple[ClickHouseType, ...], query_id: str = None, summary: Dict[str, Any] = None):
        self.result_set = result_set
        self.column_names = column_names
        self.column_types = column_types
        self.query_id = query_id
        self.summary = summary

    def named_results(self):
        for row in self.result_set:
            yield dict(zip(self.column_names, row))


class DataResult(NamedTuple):
    """
    Wrapper class for data return values and metadata at the lowest level
    """
    result: Sequence[Sequence[Any]]
    column_names: Tuple[str]
    column_types: Tuple[ClickHouseType]


local_tz = datetime.now().astimezone().tzinfo
BS = '\\'
must_escape = (BS, '\'')


def quote_identifier(identifier: str):
    first_char = identifier[0]
    if first_char in ('`', '"') and identifier[-1] == first_char:
        # Identifier is already quoted, assume that it's valid
        return identifier
    return f'`{identifier}`'


def finalize_query(query: str, parameters: Optional[Union[Sequence, Dict[str, Any]]],
                   server_tz: Optional[tzinfo] = None) -> str:
    if not parameters:
        return query
    if hasattr(parameters, 'items'):
        return query % {k: format_query_value(v, server_tz) for k, v in parameters.items()}
    return query % tuple(format_query_value(v) for v in parameters)


# pylint: disable=too-many-return-statements
def format_query_value(value: Any, server_tz: tzinfo = pytz.UTC):
    """
    Format Python values in a ClickHouse query
    :param value: Python object
    :param server_tz: Server timezone for adjusting datetime values
    :return: Literal string for python value
    """
    if value is None:
        return 'NULL'
    if isinstance(value, str):
        return f"'{''.join(f'{BS}{c}' if c in must_escape else c for c in value)}'"
    if isinstance(value, datetime):
        if value.tzinfo is None and server_tz != local_tz:
            value = value.replace(tzinfo=server_tz)
        return f"'{value.strftime('%Y-%m-%d %H:%M:%S')}'"
    if isinstance(value, date):
        return f"'{value.isoformat()}'"
    if isinstance(value, list):
        return f"[{', '.join(format_query_value(x, server_tz) for x in value)}]"
    if isinstance(value, tuple):
        return f"({', '.join(format_query_value(x, server_tz) for x in value)})"
    if isinstance(value, dict):
        pairs = [format_query_value(k, server_tz) + ':' + format_query_value(v, server_tz) for k, v in value.items()]
        return f"{{{', '.join(pairs)}}}"
    if isinstance(value, Enum):
        return format_query_value(value.value, server_tz)
    if isinstance(value, (uuid.UUID, ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return f"'{value}'"
    return str(value)


comment_re = re.compile(r"(\".*?\"|\'.*?\')|(/\*.*?\*/|(--\s)[^\n]*$)", re.MULTILINE | re.DOTALL)


def remove_sql_comments(sql: str) -> str:
    """
    Remove SQL comments.  This is useful to determine the type of SQL query, such as SELECT or INSERT, but we
    don't fully trust it to correctly ignore weird quoted strings, and other edge cases, so we always pass the
    original SQL to ClickHouse (which uses a full-fledged AST/ token parser)
    :param sql:  SQL query
    :return: SQL Query without SQL comments
    """
    def replacer(match):
        # if the 2nd group (capturing comments) is not None, it means we have captured a
        # non-quoted, actual comment string, so return nothing to remove the comment
        if match.group(2):
            return ''
        # Otherwise we've actually captured a quoted string, so return it
        return match.group(1)

    return comment_re.sub(replacer, sql)


def np_result(result: QueryResult) -> 'np.array':
    """
    Convert QueryResult to a numpy array
    :param result: QueryResult from driver
    :return: Two dimensional numpy array from result
    """
    check_numpy()
    np_types = [(name, ch_type.np_type) for name, ch_type in zip(result.column_names, result.column_types)]
    return np.array(result.result_set, dtype=np_types)


def to_pandas_df(result: QueryResult) -> 'pa.DataFrame':
    """
    Convert QueryResult to a pandas dataframe
    :param result: QueryResult from driver
    :return: Two dimensional pandas dataframe from result
    """
    check_pandas()
    return pa.DataFrame(np_result(result))


def from_pandas_df(df: 'pa.DataFrame'):
    """
    Wrap a pandas dataframe in a dictionary for use as insert keyword parameters
    :param df: Pandas data frame for insert
    :return: Simple dictionary to use for client insert function keywords
    """
    check_pandas()
    return {'column_names': df.columns, 'data': df.to_numpy()}


def to_arrow(content: bytes):
    check_arrow()
    reader = pyarrow.ipc.RecordBatchFileReader(content)
    return reader.read_all()


def arrow_buffer(table: 'pyarrow.Table') -> Tuple[Sequence[str], bytes]:
    check_arrow()
    sink = pyarrow.BufferOutputStream()
    with pyarrow.RecordBatchFileWriter(sink, table.schema) as writer:
        writer.write(table)
    return table.schema.names, sink.getvalue()



import logging

from typing import Any, Sequence

from clickhouse_connect.datatypes import registry
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.common import read_leb128, read_leb128_str
from clickhouse_connect.driver.exceptions import InterfaceError
from clickhouse_connect.driver.query import DataResult
from clickhouse_connect.driver.transform import DataTransform, QueryContext

logger = logging.getLogger(__name__)


class RowBinaryTransform(DataTransform):

    def _transform_response(self, source: Sequence, context: QueryContext) -> DataResult:
        if not isinstance(source, memoryview):
            source = memoryview(source)
        response_size = len(source)
        loc = 0
        num_columns, loc = read_leb128(source, loc)
        names = []
        for _ in range(num_columns):
            name, loc = read_leb128_str(source, loc)
            names.append(name)
        col_types = []
        for _ in range(num_columns):
            col_type, loc = read_leb128_str(source, loc)
            try:
                col_types.append(registry.get_from_name(col_type))
            except KeyError:
                raise InterfaceError(f'Unknown ClickHouse type returned for type {col_type}') from None
        convs = tuple(t.from_row_binary for t in col_types)
        result = []
        while loc < response_size:
            row = []
            for conv in convs:
                v, loc = conv(source, loc)
                row.append(v)
            result.append(row)
        return DataResult(result, tuple(names), tuple(col_types))

    def build_insert(self, data: Sequence[Sequence[Any]], *, column_type_names: Sequence[str] = None,
                     column_types: Sequence[ClickHouseType] = None, column_oriented: bool = False, **_):
        if not column_types:
            column_types = [registry.get_from_name(name) for name in column_type_names]
        convs = tuple(t.to_row_binary for t in column_types)
        if column_oriented:
            data = tuple(zip(*data))
        output = bytearray()
        for row in data:
            for (value, conv) in zip(row, convs):
                conv(value, output)
        return output



from typing import Optional, Union

from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.httpclient import HttpClient


# pylint: disable=too-many-arguments
def create_client(host: str = 'localhost', username: str = None, password: str = '', database: str = '__default__',
                  interface: Optional[str] = None, port: int = 0, secure: Union[bool, str] = False,
                  **kwargs) -> Client:
    use_tls = str(secure).lower() == 'true' or interface == 'https' or (not interface and port in (443, 8443))
    if not interface:
        interface = 'https' if use_tls else 'http'
    port = port or default_port(interface, use_tls)
    if username is None and 'user' in kwargs:
        username = kwargs.pop('user')
    if password and username is None:
        username = 'default'
    if interface.startswith('http'):
        cc_client = HttpClient(interface, host, port, username, password, database, **kwargs)
    else:
        raise ProgrammingError(f'Unrecognized client type {interface}')
    return cc_client


def default_port(interface: str, secure: bool):
    if interface.startswith('http'):
        return 8443 if secure else 8123
    raise ValueError('Unrecognized ClickHouse interface')



from clickhouse_connect.dbapi.connection import Connection


apilevel = '2.0'         # PEP 249  DB API level
threadsafety = 2         # PEP 249  Threads may share the module and connections.
paramstyle = 'pyformat'  # PEP 249  Python extended format codes, e.g. ...WHERE name=%(name)s


class Error(Exception):
    pass


def connect(**kwargs):
    return Connection(**kwargs)



import logging
from typing import Dict, Type

from sqlalchemy.exc import CompileError

from clickhouse_connect.datatypes.base import ClickHouseType, TypeDef, EMPTY_TYPE_DEF
from clickhouse_connect.datatypes.registry import parse_name, type_map
from clickhouse_connect.driver.query import format_query_value

logger = logging.getLogger(__name__)


class ChSqlaType:
    """
    A SQLAlchemy TypeEngine that wraps a ClickHouseType.  We don't extend TypeEngine directly, instead all concrete
    subclasses will inherit from TypeEngine.
    """
    ch_type: ClickHouseType = None
    generic_type: None
    _ch_type_cls = None
    _instance = None
    _instance_cache: Dict[TypeDef, 'ChSqlaType'] = None

    def __init_subclass__(cls):
        """
        Registers ChSqla type in the type map and sets the underlying ClickHouseType class to use to initialize
        ChSqlaType instances
        """
        base = cls.__name__
        if not cls._ch_type_cls:
            try:
                cls._ch_type_cls = type_map[base]
            except KeyError:
                logger.warning('Attempted to register SQLAlchemy type without corresponding ClickHouse Type')
                return
        schema_types.append(base)
        sqla_type_map[base] = cls
        cls._instance_cache = {}

    @classmethod
    def build(cls, type_def: TypeDef):
        """
        Factory function for building a ChSqlaType based on the type definition
        :param type_def: -- TypeDef tuple that defines arguments for this instance
        :return: Shared instance of a configured ChSqlaType
        """
        return cls._instance_cache.setdefault(type_def, cls(type_def=type_def))

    def __init__(self, type_def: TypeDef = EMPTY_TYPE_DEF):
        """
        Basic constructor that does nothing but set the wrapped ClickHouseType.  It is overridden in some cases
        to add specific SqlAlchemy behavior when constructing subclasses "by hand", in which case the type_def
        parameter is normally set to None and other keyword parameters used for construction
        :param type_def: TypeDef tuple used to build the underlying ClickHouseType.  This is normally populated by the
        parse_name function
        """
        self.type_def = type_def
        self.ch_type = self._ch_type_cls.build(type_def)

    @property
    def name(self):
        return self.ch_type.name

    @name.setter
    def name(self, name):  # Keep SQLAlchemy from overriding our ClickHouse name
        pass

    @property
    def nullable(self):
        return self.ch_type.nullable

    @property
    def low_card(self):
        return self.ch_type.low_card

    @staticmethod
    def result_processor():
        """
        Override for the SqlAlchemy TypeEngine result_processor method, which is used to convert row values to the
        correct Python type.  The core driver handles this automatically, so we always return None.
        """
        return None

    @staticmethod
    def _cached_result_processor(*_):
        """
        Override for the SqlAlchemy TypeEngine _cached_result_processor method to prevent weird behavior
        when SQLAlchemy tries to cache.
        """
        return None

    @staticmethod
    def _cached_literal_processor(*_):
        """
        Override for the SqlAlchemy TypeEngine _cached_literal_processor. We delegate to the driver format_query_value
        method and should be able to ignore literal_processor definitions in the dialect, which are verbose and
        confusing.
        """
        return format_query_value

    def _compiler_dispatch(self, _visitor, **_):
        """
        Override for the SqlAlchemy TypeEngine _compiler_dispatch method to sidestep unnecessary layers and complexity
        when generating the type name.  The underlying ClickHouseType generates the correct name
        :return: Name generated by the underlying driver.
        """
        return self.name


sqla_type_map: Dict[str, Type[ChSqlaType]] = {}
schema_types = []


def sqla_type_from_name(name: str) -> ChSqlaType:
    """
    Factory function to convert a ClickHouse type name to the appropriate ChSqlaType
    :param name: Name returned from ClickHouse using Native protocol or WithNames format
    :return: ChSqlaType
    """
    base, name, type_def = parse_name(name)
    try:
        type_cls = sqla_type_map[base]
    except KeyError:
        err_str = f'Unrecognized ClickHouse type base: {base} name: {name}'
        logger.error(err_str)
        raise CompileError(err_str) from KeyError
    return type_cls.build(type_def)



# pylint: disable=unused-import
from clickhouse_connect.driver.exceptions import NotSupportedError

HAS_NUMPY = True
HAS_PANDAS = True
HAS_ARROW = True


try:
    import numpy as np
except ImportError:
    HAS_NUMPY = False

try:
    import pandas as pa
except ImportError:
    HAS_PANDAS = False

try:
    import pyarrow as arrow
except ImportError:
    HAS_ARROW = False


def check_numpy():
    if not HAS_NUMPY:
        raise NotSupportedError("Numpy package is not installed")


def check_pandas():
    if not HAS_PANDAS:
        raise NotSupportedError("Pandas package is not installed")


def check_arrow():
    if not HAS_ARROW:
        raise NotSupportedError("PyArrow package is not installed")



import pytz

from datetime import date, timedelta, datetime
from struct import unpack_from as suf, pack as sp
from typing import Union, Sequence, MutableSequence

from clickhouse_connect.datatypes.base import TypeDef, ArrayType
from clickhouse_connect.driver.common import read_uint64, array_column, write_array

epoch_start_date = date(1970, 1, 1)


class Date(ArrayType):
    _array_type = 'H'
    python_null = epoch_start_date
    np_type = 'M8[D]'
    python_type = date

    def _from_row_binary(self, source: Sequence, loc: int):
        return epoch_start_date + timedelta(suf('<H', source, loc)[0]), loc + 2

    def _to_row_binary(self, value: date, dest: MutableSequence):
        dest += sp('<H', (value - epoch_start_date).days, )

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        return [epoch_start_date + timedelta(days) for days in column], loc

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        esd = epoch_start_date
        if self.nullable:
            write_array(self._array_type, [0 if x is None else (x - esd).days for x in column], dest)
        else:
            write_array(self._array_type, [(x - esd).days for x in column], dest)


class Date32(Date):
    _array_type = 'i'

    def _from_row_binary(self, source: Sequence, loc: int):
        return epoch_start_date + timedelta(suf('<i', source, loc)[0]), loc + 4

    def _to_row_binary(self, value: date, dest: MutableSequence):
        dest += (value - epoch_start_date).days.to_bytes(4, 'little', signed=True)


from_ts_naive = datetime.utcfromtimestamp
from_ts_tz = datetime.fromtimestamp


# pylint: disable=abstract-method
class DateTime(ArrayType):
    __slots__ = ('_from_row_binary',)
    _array_type = 'I'
    np_type = 'M8[us]'
    python_null = from_ts_naive(0)
    python_type = datetime

    def __init__(self, type_def: TypeDef):
        if type_def.values:
            tzinfo = pytz.timezone(type_def.values[0][1:-1])
            self._from_row_binary = lambda source, loc: (from_ts_tz(suf('<L', source, loc)[0], tzinfo), loc + 4)
        else:
            self._from_row_binary = lambda source, loc: (from_ts_naive(suf('<L', source, loc)[0]), loc + 4)
        super().__init__(type_def)

    def _to_row_binary(self, value: datetime, dest: MutableSequence):
        dest += sp('<I', int(value.timestamp()), )

    def _read_native_binary(self, source: Sequence, loc: int, num_rows: int):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        fts = from_ts_naive
        return [fts(ts) for ts in column], loc

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        if self.nullable:
            column = [int(x.timestamp()) if x else 0 for x in column]
        else:
            column = [int(x.timestamp()) for x in column]
        write_array(self._array_type, column, dest)


class DateTime64(ArrayType):
    __slots__ = 'prec', 'tzinfo'
    _array_type = 'Q'
    python_null = epoch_start_date
    python_type = datetime

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str
        self.prec = 10 ** type_def.values[0]
        if len(type_def.values) > 1:
            self.tzinfo = pytz.timezone(type_def.values[1][1:-1])
            self._read_native_binary = self._read_native_tz
        else:
            self._read_native_binary = self._read_native_naive
            self.tzinfo = None

    def _from_row_binary(self, source, loc):
        ticks, loc = read_uint64(source, loc)
        seconds = ticks // self.prec
        dt_sec = datetime.fromtimestamp(seconds, self.tzinfo)
        microseconds = ((ticks - seconds * self.prec) * 1000000) // self.prec
        return dt_sec + timedelta(microseconds=microseconds), loc

    def _to_row_binary(self, value: datetime, dest: bytearray):
        microseconds = int(value.timestamp()) * 1000000 + value.microsecond
        dest += (int(microseconds * self.prec) // 1000000).to_bytes(8, 'little', signed=True)

    def _read_native_tz(self, source: Sequence, loc: int, num_rows: int):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        new_col = []
        app = new_col.append
        dt_from = datetime.fromtimestamp
        prec = self.prec
        tz_info = self.tzinfo
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds, tz_info)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col, loc

    def _read_native_naive(self, source: Sequence, loc: int, num_rows: int):
        column, loc = array_column(self._array_type, source, loc, num_rows)
        new_col = []
        app = new_col.append
        dt_from = datetime.utcfromtimestamp
        prec = self.prec
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col, loc

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        prec = self.prec
        if self.nullable:
            column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 if x else 0 for x in column]
        else:
            column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 for x in column]
        write_array(self._array_type, column, dest)



from typing import NamedTuple

from clickhouse_connect.datatypes.registry import get_from_name


class ColumnDef(NamedTuple):
    """
    ClickHouse column definition from DESCRIBE TABLE command
    """
    name: str
    type: str
    default_type: str
    default_expression: str
    comment: str
    codec_expression: str
    ttl_expression: str

    @property
    def ch_type(self):
        return get_from_name(self.type)


class SettingDef(NamedTuple):
    """
    ClickHouse setting definition from system.settings table
    """
    name: str
    value: str
    changed: int
    description: str
    readonly: int
    type: str



from sqlalchemy.sql.ddl import DDL
from sqlalchemy.exc import ArgumentError

from clickhouse_connect.driver.query import quote_identifier


class CreateDatabase(DDL):
    """
    SqlAlchemy DDL statement that is essentially an alternative to the built in CreateSchema DDL class
    """
    # pylint: disable-msg=too-many-arguments
    def __init__(self, name: str, engine: str = None, zoo_path: str = None, shard_name: str = '{shard}',
                 replica_name: str = '{replica}'):
        """
        :param name: Database name
        :param engine: Database ClickHouse engine type
        :param zoo_path: ClickHouse zookeeper path for Replicated database engine
        :param shard_name: Clickhouse shard name for Replicated database engine
        :param replica_name: Replica name for Replicated database engine
        """
        if engine and engine not in ('Ordinary', 'Atomic', 'Lazy', 'Replicated'):
            raise ArgumentError(f'Unrecognized engine type {engine}')
        stmt = f'CREATE DATABASE {quote_identifier(name)}'
        if engine:
            stmt += f' Engine {engine}'
            if engine == 'Replicated':
                if not zoo_path:
                    raise ArgumentError('zoo_path is required for Replicated Database Engine')
                stmt += f" ('{zoo_path}', '{shard_name}', '{replica_name}'"
        super().__init__(stmt)


class DropDatabase(DDL):
    """
    Alternative DDL statement for built in SqlAlchemy DropSchema DDL class
    """
    def __init__(self, name: str):
        super().__init__(f'DROP DATABASE {quote_identifier(name)}')



import pkg_resources

from clickhouse_connect.driver import create_client
from clickhouse_connect.entry_points import validate_entrypoints

driver_name = 'clickhousedb'


def get_client(**kwargs):
    return create_client(**kwargs)


def check_ep():
    assert validate_entrypoints() == 0


def version():
    return pkg_resources.get_distribution('clickhouse-connect').version



import logging
import re

from typing import Optional, Sequence

from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.driver.common import unescape_identifier
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver import Client
from clickhouse_connect.driver.parser import parse_callable
from clickhouse_connect.driver.query import remove_sql_comments

logger = logging.getLogger(__name__)

query_re = re.compile(r'^\s*(SELECT|SHOW|DESCRIBE|WITH|EXISTS)\s', re.IGNORECASE)
insert_re = re.compile(r'^\s*INSERT\s+INTO\s+(.*$)', re.IGNORECASE)
str_type = get_from_name('String')
int_type = get_from_name('Int32')


class Cursor:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """
    def __init__(self, client: Client):
        self.client = client
        self.arraysize = 1
        self.data: Optional[Sequence] = None
        self.names = []
        self.types = []
        self._rowcount = 0
        self._ix = 0

    def check_valid(self):
        if self.data is None:
            raise ProgrammingError('Cursor is not valid')

    @property
    def description(self):
        return [(n, t.name, None, None, None, None, True) for n, t in zip(self.names, self.types)]

    @property
    def rowcount(self):
        return self._rowcount

    def close(self):
        self.data = None

    def execute(self, operation:str, parameters=None):
        if query_re.match(remove_sql_comments(operation)):
            query_result = self.client.query(operation, parameters)
            self.data = query_result.result_set
            self.names = query_result.column_names
            self.types = query_result.column_types
        else:
            v = self.client.command(operation, parameters)
            self.types = [str_type]
            self.names = ['response']
            if not v:
                v = 'OK'
            elif isinstance(v, int):
                self.types = [int_type]
            if isinstance(v, list):
                v = '\t'.join(v)
            self.data = [[v]]
        self._rowcount = len(self.data)

    def _try_bulk_insert(self, operation:str, data):
        match = insert_re.match(remove_sql_comments(operation))
        if not match:
            return False
        temp = match.group(1)
        table_end = min(temp.find(' '), temp.find('('))
        table = temp[:table_end].strip()
        temp = temp[table_end:].strip()
        if temp[0] == '(':
            _, op_columns, temp = parse_callable(temp)
        else:
            op_columns = None
        if 'VALUES' not in temp.upper():
            return False
        col_names = list(data[0].keys())
        if op_columns and {unescape_identifier(x) for x in op_columns} != set(col_names):
            return False  # Data sent in doesn't match the columns in the insert statement
        if self.client.column_inserts:
            data_values = [[row[k] for row in data] for k in col_names]
        else:
            data_values = [list(row.values()) for row in data]
        self.client.insert(table, data_values, col_names, column_oriented=self.client.column_inserts)
        self.data = []
        return True

    def executemany(self, operation, parameters):
        if not parameters or self._try_bulk_insert(operation, parameters):
            return
        self.data = []
        try:
            for param_row in parameters:
                query_result = self.client.query(operation, param_row)
                self.data.extend(query_result.result_set)
                if self.names or self.types:
                    if query_result.column_names != self.names:
                        logger.warning('Inconsistent column names %s : %s for operation %s in cursor executemany',
                                        self.names, query_result.column_names, operation)
                else:
                    self.names = query_result.column_names
                    self.types = query_result.column_types
        except TypeError as ex:
            raise ProgrammingError(f'Invalid parameters {parameters} passed to cursor executemany') from ex
        self._rowcount = len(self.data)

    def fetchall(self):
        self.check_valid()
        ret = self.data
        self._ix = self._rowcount
        return ret

    def fetchone(self):
        self.check_valid()
        if self._ix >= self._rowcount:
            return None
        val = self.data[self._ix]
        self._ix += 1
        return val

    def fetchmany(self, size: int = -1):
        self.check_valid()
        end = self._ix + max(size, self._rowcount - self._ix)
        ret = self.data[self._ix: end]
        self._ix = end
        return ret

    def nextset(self):
        raise NotImplementedError

    def callproc(self, *args, **kwargs):
        raise NotImplementedError



import logging
import json
import atexit
import re
import http as PyHttp

from typing import Optional, Dict, Any, Sequence, Union, List
from requests import Session, Response, get as req_get
from requests.exceptions import RequestException

from clickhouse_connect.datatypes import registry
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.exceptions import DatabaseError, OperationalError, ProgrammingError
from clickhouse_connect.driver.httpadapter import KeepAliveAdapter
from clickhouse_connect.driver.native import NativeTransform
from clickhouse_connect.driver.query import QueryResult, DataResult, QueryContext, finalize_query, quote_identifier
from clickhouse_connect.driver.rowbinary import RowBinaryTransform

logger = logging.getLogger(__name__)
columns_only_re = re.compile(r'LIMIT 0\s*$', re.IGNORECASE)

# Create a single HttpAdapter that will be shared by all client sessions.  This is intended to make
# the client as thread safe as possible while sharing a single connection pool.  For the same reason we
# don't call the Session.close() method from the client so the connection pool remains available
http_adapter = KeepAliveAdapter(pool_connections=4, pool_maxsize=8, max_retries=0)
atexit.register(http_adapter.close)

# Increase this number just to be safe when ClickHouse is returning progress headers
PyHttp._MAXHEADERS = 10000  # pylint: disable=protected-access


# pylint: disable=too-many-instance-attributes
class HttpClient(Client):
    valid_transport_settings = {'database', 'buffer_size', 'session_id', 'compress', 'decompress',
                                'session_timeout', 'session_check', 'query_id', 'quota_key', 'wait_end_of_query'}

    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
    def __init__(self,
                 interface: str,
                 host: str,
                 port: int,
                 username: str,
                 password: str,
                 database: str,
                 compress: bool = True,
                 data_format: str = 'native',
                 query_limit: int = 5000,
                 query_retries: int = 2,
                 connect_timeout: int = 10,
                 send_receive_timeout: int = 300,
                 client_name: str = 'clickhouse-connect',
                 send_progress: bool = True,
                 verify: bool = True,
                 ca_cert: str = None,
                 client_cert: str = None,
                 client_cert_key: str = None,
                 **kwargs):
        """
        Create an HTTP ClickHouse Connect client
        :param interface: http or https
        :param host: hostname
        :param port: host port
        :param username: ClickHouse user
        :param password: ClickHouse password
        :param database: Default database for the connection
        :param compress: Accept compressed HTTP type from server (brotli or gzip)
        :param data_format: Dataformat -- either 'native' or 'rb' (RowBinary)
        :param query_limit: Default LIMIT on returned rows
        :param connect_timeout:  Timeout in seconds for the http connection
        :param send_receive_timeout: Read timeout in seconds for http connection
        :param client_name: Http user agent header value
        :param send_progress: Ask ClickHouse to send progress headers.  Used for summary and keep alive
        :param verify: Verify the server certificate in secure/https mode
        :param ca_cert: If verify is True, the file path to Certificate Authority root to validate ClickHouse server
         certificate, in .pem format.  Ignored if verify is False.  This is not necessary if the ClickHouse server
         certificate is a globally trusted root as verified by the operating system
        :param client_cert: File path to a TLS Client certificate in .pem format.  This file should contain any
          applicable intermediate certificates
        :param client_cert_key: File path to the private key for the Client Certificate.  Required if the private key
          is not included the Client Certificate key file
        :param kwargs: Optional clickhouse setting values (str/value) for every connection request
        """
        self.url = f'{interface}://{host}:{port}'
        session = Session()
        session.stream = False
        session.max_redirects = 3
        if client_cert:
            if not username:
                raise ProgrammingError('username parameter is required for Mutual TLS authentication')
            if client_cert_key:
                session.cert = (client_cert, client_cert_key)
            else:
                session.cert = client_cert
            session.headers['X-ClickHouse-User'] = username
            session.headers['X-ClickHouse-SSL-Certificate-Auth'] = 'on'
        else:
            session.auth = (username, password if password else '') if username else None
        session.verify = False
        if interface == 'https':
            if verify and ca_cert:
                session.verify = ca_cert
            else:
                session.verify = verify

        # Remove the default session adapters, they are not used and this avoids issues with their connection pools
        session.adapters.pop('http://').close()
        session.adapters.pop('https://').close()
        session.mount(self.url, adapter=http_adapter)
        session.headers['User-Agent'] = client_name

        if data_format == 'native':
            self.read_format = self.write_format = 'Native'
            self.column_inserts = True
            self.transform = NativeTransform()
        elif data_format in ('row_binary', 'rb'):
            self.read_format = 'RowBinaryWithNamesAndTypes'
            self.write_format = 'RowBinary'
            self.column_inserts = False
            self.transform = RowBinaryTransform()
        self.session = session
        self.connect_timeout = connect_timeout
        self.read_timeout = send_receive_timeout
        self.query_retries = query_retries
        settings = kwargs.copy()
        if send_progress:
            settings['send_progress_in_http_headers'] = '1'
            settings['wait_end_of_query'] = '1'
            if self.read_timeout > 10:
                progress_interval = (self.read_timeout - 5) * 1000
            else:
                progress_interval = 120000  # Two minutes
            settings['http_headers_progress_interval_ms'] = str(progress_interval)
        if compress:
            session.headers['Accept-Encoding'] = 'gzip'
            settings['enable_http_compression'] = '1'
        super().__init__(database=database, query_limit=query_limit, uri=self.url)
        self.session.params = self._validate_settings(settings, True)

    def client_setting(self, name, value):
        if isinstance(value, bool):
            value = '1' if value else '0'
        self.session.params[name] = str(value)

    def _prep_query(self, context: QueryContext):
        final_query = super()._prep_query(context)
        if context.is_insert:
            return final_query
        return f'{final_query}\n FORMAT {self.write_format}'

    def _query_with_context(self, context: QueryContext) -> QueryResult:
        headers = {'Content-Type': 'text/plain; charset=utf-8'}
        params = {'database': self.database}
        params.update(self._validate_settings(context.settings, True))
        if columns_only_re.search(context.uncommented_query):
            response = self._raw_request(f'{context.final_query}\n FORMAT JSON',
                                         params, headers, retries=self.query_retries)
            json_result = json.loads(response.content)
            # ClickHouse will respond with a JSON object of meta, data, and some other objects
            # We just grab the column names and column types from the metadata sub object
            names: List[str] = []
            types: List[ClickHouseType] = []
            for col in json_result['meta']:
                names.append(col['name'])
                types.append(registry.get_from_name(col['type']))
            data_result = DataResult([], tuple(names), tuple(types))
        else:
            response = self._raw_request(self._prep_query(context), params, headers, retries=self.query_retries)
            data_result = self.transform.parse_response(response.content, context)
        summary = {}
        if 'X-ClickHouse-Summary' in response.headers:
            try:
                summary = json.loads(response.headers['X-ClickHouse-Summary'])
            except json.JSONDecodeError:
                pass
        return QueryResult(data_result.result, data_result.column_names, data_result.column_types,
                           response.headers.get('X-ClickHouse-Query-Id'), summary)

    def data_insert(self,
                    table: str,
                    column_names: Sequence[str],
                    data: Sequence[Sequence[Any]],
                    column_types: Sequence[ClickHouseType],
                    settings: Optional[Dict[str, Any]] = None,
                    column_oriented: bool = False):
        """
        See BaseClient doc_string for this method
        """
        insert_block = self.transform.build_insert(data, column_types=column_types, column_names=column_names,
                                                   column_oriented=column_oriented)
        self.raw_insert(table, column_names, insert_block, settings, self.write_format)

    def raw_insert(self, table: str,
                   column_names: Sequence[str],
                   insert_block: Union[str, bytes],
                   settings: Optional[Dict] = None,
                   fmt: Optional[str] = None):
        """
        See BaseClient doc_string for this method
        """
        column_ids = [quote_identifier(x) for x in column_names]
        write_format = fmt if fmt else self.write_format
        headers = {'Content-Type': 'application/octet-stream'}
        params = {'query': f"INSERT INTO {table} ({', '.join(column_ids)}) FORMAT {write_format}",
                  'database': self.database}
        if isinstance(insert_block, str):
            insert_block = insert_block.encode()
        params.update(self._validate_settings(settings, True))
        response = self._raw_request(insert_block, params, headers)
        logger.debug('Insert response code: %d, content: %s', response.status_code, response.content)

    def command(self,
                cmd,
                parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                data: Union[str, bytes] = None,
                settings: Optional[Dict] = None,
                use_database: int = True) -> Union[str, int, Sequence[str]]:
        """
        See BaseClient doc_string for this method
        """
        cmd = finalize_query(cmd, parameters, self.server_tz)
        headers = {}
        params = {}
        payload = None
        if isinstance(data, str):
            headers['Content-Type'] = 'text/plain; charset=utf-8'
            payload = data.encode()
        elif isinstance(data, bytes):
            headers['Content-Type'] = 'application/octet-stream'
            payload = data
        if payload is None:
            if not cmd:
                raise ProgrammingError('Command sent without query or recognized data')
            payload = cmd
        elif cmd:
            params['query'] = cmd
        if use_database:
            params['database'] = self.database
        params.update(self._validate_settings(settings, True))
        method = 'POST' if payload else 'GET'
        response = self._raw_request(payload, params, headers, method)
        result = response.content.decode('utf8')[:-1].split('\t')
        if len(result) == 1:
            try:
                return int(result[0])
            except ValueError:
                return result[0]
        return result

    def _raw_request(self,
                     data,
                     params: Dict[str, Any],
                     headers: Optional[Dict[str, Any]] = None,
                     method: str = 'POST',
                     retries: int = 0) -> Response:
        if isinstance(data, str):
            data = data.encode()
        attempts = 0
        while True:
            attempts += 1
            try:
                response: Response = self.session.request(method, self.url,
                                                          headers=headers,
                                                          timeout=(self.connect_timeout, self.read_timeout),
                                                          data=data,
                                                          params=params)
            except RequestException as ex:
                logger.exception('Unexpected Http Driver Exception')
                raise OperationalError(f'Error executing HTTP request {self.url}') from ex
            if 200 <= response.status_code < 300:
                return response
            err_str = f'HTTPDriver url {self.url} returned response code {response.status_code})'
            logger.error(err_str)
            if response.content:
                err_msg = response.content.decode(errors='backslashreplace')
                logger.error(str(err_msg))
                err_str = f':{err_str}\n {err_msg[0:240]}'
            if response.status_code not in (429, 503, 504):
                raise DatabaseError(err_str)
            if attempts > retries:
                raise OperationalError(err_str)

    def ping(self):
        """
        See BaseClient doc_string for this method
        """
        try:
            response = req_get(f'{self.url}/ping', timeout=3)
            return 200 <= response.status_code < 300
        except RequestException:
            logger.debug('ping failed', exc_info=True)
            return False

    def raw_query(self,
                  query: str,
                  parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  fmt: str = None):
        """
        See BaseClient doc_string for this method
        """
        final_query = finalize_query(query, parameters, self.server_tz)
        if fmt:
            final_query += f'\n FORMAT {fmt}'
        return self._raw_request(final_query, self._validate_settings(settings, True)).content


def reset_connections():
    """
    Used for tests to force new connection by resetting the singleton HttpAdapter
    """

    global http_adapter  # pylint: disable=global-statement
    http_adapter = KeepAliveAdapter(pool_connections=4, pool_maxsize=8, max_retries=0)



from urllib.parse import urlparse, parse_qs

from clickhouse_connect.dbapi.cursor import Cursor
from clickhouse_connect.driver import create_client

# pylint: disable=too-many-arguments
from clickhouse_connect.driver.query import QueryResult


class Connection:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """
    def __init__(self, dsn: str = None, username: str = None, password: str = None, host: str = None,
                 database: str = None, interface: str = None, port: int = 0, **kwargs):
        settings = kwargs.copy()
        if dsn:
            parsed = urlparse(dsn)
            username = username or parsed.username
            password = password or parsed.password
            host = host or parsed.hostname
            port = port or parsed.port
            if parsed.path and not database:
                database = parsed.path[1:].split('/')[0]
            database = database or parsed.path
            settings = dict(parse_qs(parsed.query)).update(settings)
        self.client = create_client(host, username, password, database, interface, port, **settings)
        self.timezone = self.client.server_tz

    def close(self):
        self.client.close()

    def commit(self):
        pass

    def rollback(self):
        pass

    def command(self, cmd: str):
        return self.client.command(cmd)

    def raw_query(self, query: str) -> QueryResult:
        return self.client.query(query)

    def cursor(self):
        return Cursor(self.client)



import os
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError
from setuptools import setup, find_packages, Extension

c_modules = []

try:
    from Cython.Build import cythonize
    print ('Using Cython to build cython modules')
    c_modules = cythonize('clickhouse_connect/driverc/*.pyx')
    C_PKG = None
except ImportError:
    print ('Cython Not Successfully Installed, Attempt to compile c files directly')
    cythonize = None
    C_PKG = 'clickhouse_connect/driverc'
    c_modules = [Extension('creaders', ['clickhouse_connect/driverc/creaders.c'])]


def run_setup(try_c: bool = True):
    if try_c:
        kwargs = {'ext_modules': c_modules}
        if C_PKG:
            kwargs['ext_package'] = C_PKG
    else:
        kwargs = {}

    project_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(project_dir, 'README.md'), encoding='utf-8') as read_me:
        long_desc = read_me.read()

    version_fn = '.dev_version' if os.path.isfile('.dev_version') else 'clickhouse_connect/VERSION'
    with open(os.path.join(project_dir, version_fn), encoding='utf-8') as version_file:
        version = version_file.readline()

    setup(
        name='clickhouse-connect',
        author='ClickHouse Inc.',
        author_email='clients@clickhouse.com',
        description='ClickHouse core driver, SqlAlchemy, and Superset libraries',
        version=version,
        long_description=long_desc,
        long_description_content_type='text/markdown',
        url='https://github.com/ClickHouse/clickhouse-connect',
        packages=find_packages(exclude=['tests*']),
        python_requires='~=3.7',
        license='Apache License 2.0',
        install_requires=[
            'requests',
            'pytz'
        ],
        extras_require={
            'sqlalchemy': ['sqlalchemy>1.3.21, <1.4'],
            'superset': ['apache_superset>=1.4.1', 'sqlalchemy>1.3.21, <1.4'],
            'numpy': ['numpy'],
            'pandas': ['pandas'],
            'arrow': ['pyarrow'],
            'orjson': ['orjson']
        },
        entry_points={
            'sqlalchemy.dialects': ['clickhousedb.connect=clickhouse_connect.cc_sqlalchemy.dialect:ClickHouseDialect',
                                    'clickhousedb=clickhouse_connect.cc_sqlalchemy.dialect:ClickHouseDialect'],
            'superset.db_engine_specs': ['clickhousedb=clickhouse_connect.cc_superset.engine:ClickHouseEngineSpec']
        },
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10'
        ],
        **kwargs
    )


try:
    run_setup()
except (CCompilerError, DistutilsExecError, DistutilsPlatformError, IOError, SystemExit) as e:
    print(f'Unable to compile C extensions for faster performance due to {e}, will use pure Python')
    run_setup(False)



import array
import sys

from typing import Tuple, Sequence, MutableSequence

# pylint: disable=invalid-name
must_swap = sys.byteorder == 'big'
int_size = array.array('i').itemsize
low_card_version = 1

array_map = {1: 'b', 2: 'h', 4: 'i', 8: 'q'}
decimal_prec = {32: 9, 64: 18, 128: 38, 256: 79}

if int_size == 2:
    array_map[4] = 'l'

array_sizes = {v: k for k, v in array_map.items()}
array_sizes['f'] = 4
array_sizes['d'] = 8


def array_type(size: int, signed: bool):
    """
    Determines the Python array.array code for the requested byte size
    :param size: byte size
    :param signed: whether int types should be signed or unsigned
    :return: Python array.array code
    """
    try:
        code = array_map[size]
    except KeyError:
        return None
    return code if signed else code.upper()


def array_column(code: str, source: Sequence, loc: int, num_rows: int) -> Tuple[array.array, int]:
    """
    Read the source binary buffer into a Python array.array
    :param code: Python array.array type code
    :param source: Source byte buffer like object
    :param loc: Start read position in the buffer
    :param num_rows: Number of rows to read
    :return: array.array of the requested data type plus next buffer read location
    """
    column = array.array(code)
    sz = column.itemsize * num_rows
    end = loc + sz
    column.frombytes(source[loc: end])
    if must_swap:
        column.byteswap()
    return column, end


def write_array(code: str, column: Sequence, dest: MutableSequence):
    """
    Write a column of native Python data matching the array.array code
    :param code: Python array.array code matching the column data type
    :param column: Column of native Python values
    :param dest: Destination byte buffer
    """
    if column and not isinstance(column[0], (int, float)):
        if code in ('f', 'F', 'd', 'D'):
            column = [float(x) for x in column]
        else:
            column = [int(x) for x in column]
    buff = array.array(code, column)
    if must_swap:
        buff.byteswap()
    dest += buff.tobytes()


def read_uint64(source: Sequence, loc: int) -> Tuple[int, int]:
    """
    Read a single UInt64 value from a data buffer
    :param source: Source binary buffer
    :param loc: Source start location
    :return: UInt64 value from buffer plus new source location
    """
    return int.from_bytes(source[loc: loc + 8], 'little'), loc + 8


def write_uint64(value: int, dest: MutableSequence):
    """
    Write a single UInt64 value to a binary write buffer
    :param value: UInt64 value to write
    :param dest: Destination byte buffer
    """
    dest.extend(value.to_bytes(8, 'little'))


def read_leb128(source: Sequence, loc: int) -> Tuple[int, int]:
    """
    Read a LEB128 encoded integer value from a source buffer
    :param source: Source binary buffer
    :param loc: Start location
    :return: Leb128 value plus next read location
    """
    length = 0
    ix = 0
    while True:
        b = source[loc + ix]
        length = length + ((b & 0x7f) << (ix * 7))
        ix += 1
        if (b & 0x80) == 0:
            break
    return length, loc + ix


def read_leb128_str(source: Sequence, loc: int, encoding: str = 'utf8') -> Tuple[str, int]:
    """
    Read a LEB128 encoded string from a binary source buffer
    :param source: Binary source buffer
    :param loc: read start location
    :param encoding: expected string encoding
    :return: Complete string plus new read location in source buffer
    """
    length, loc = read_leb128(source, loc)
    return str(source[loc:loc + length], encoding), loc + length


def write_leb128(value: int, dest: MutableSequence):
    """
    Write a LEB128 encoded integer to a target binary buffer
    :param value: Integer value (positive only)
    :param dest: Target buffer
    """
    while True:
        b = value & 0x7f
        value >>= 7
        if value == 0:
            dest.append(b)
            return
        dest.append(0x80 | b)


def to_leb128(value: int) -> bytearray:
    """
    Create a byte array representing a LEB128 encoded integer
    :param value: Integer value to encode
    :return: bytearray with encoding value
    """
    result = bytearray()
    while True:
        b = value & 0x7f
        value >>= 7
        if value == 0:
            result.append(b)
            break
        result.append(0x80 | b)
    return result


def decimal_size(prec: int):
    """
    Determine the bit size of a ClickHouse or Python Decimal needed to store a value of the requested precision
    :param prec: Precision of the Decimal in total number of base 10 digits
    :return: Required bit size
    """
    if prec < 1 or prec > 79:
        raise ArithmeticError(f'Invalid precision {prec} for ClickHouse Decimal type')
    if prec < 10:
        return 32
    if prec < 19:
        return 64
    if prec < 39:
        return 128
    return 256


def unescape_identifier(x: str) -> str:
    if x.startswith('`') and x.endswith('`'):
        return x[1:-1]
    return x



import logging
from typing import Type, Sequence, Optional, Dict

from sqlalchemy.exc import ArgumentError, SQLAlchemyError
from sqlalchemy.sql.base import SchemaEventTarget
from sqlalchemy.sql.visitors import Visitable

logger = logging.getLogger(__name__)

engine_map: Dict[str, Type['TableEngine']] = {}


def tuple_expr(expr_name, value):
    """
    Create a table parameter with a tuple or list correctly formatted
    :param expr_name: parameter
    :param value: string or tuple of strings to format
    :return: formatted parameter string
    """
    if value is None:
        return ''
    v = f'{expr_name.strip()}'
    if isinstance(value, (tuple, list)):
        return f" {v} ({','.join(value)})"
    return f'{v} {value}'


class TableEngine(SchemaEventTarget, Visitable):
    """
    SqlAlchemy Schema element to support ClickHouse table engines.  At the moment provides no real
    functionality other than the CREATE TABLE argument string
    """
    arg_names = ()
    quoted_args = set()
    optional_args = set()
    eng_params = ()

    def __init_subclass__(cls, **kwargs):
        engine_map[cls.__name__] = cls

    def __init__(self, kwargs):
        # pylint: disable=no-value-for-parameter
        Visitable.__init__(self)
        self.name = self.__class__.__name__
        te_name = f'{self.name} Table Engine'
        engine_args = []
        for arg_name in self.arg_names:
            v = kwargs.pop(arg_name, None)
            if v is None:
                if arg_name in self.optional_args:
                    continue
                raise ValueError(f'Required engine parameter {arg_name} not provided for {te_name}')
            if arg_name in self.quoted_args:
                engine_args.append(f"'{v}'")
            else:
                engine_args.append(v)
        if engine_args:
            self.arg_str = f'({", ".join(engine_args)})'
        params = []
        for param_name in self.eng_params:
            v = kwargs.pop(param_name, None)
            if v is not None:
                params.append(tuple_expr(param_name.upper().replace('_', ' '), v))

        self.full_engine = 'Engine ' + self.name
        if engine_args:
            self.full_engine += f'({", ".join(engine_args)})'
        if params:
            self.full_engine += ' ' + ' '.join(params)

    def compile(self):
        return self.full_engine

    def check_primary_keys(self, primary_keys: Sequence):
        raise SQLAlchemyError(f'Table Engine {self.name} does not support primary keys')

    def _set_parent(self, parent):
        parent.engine = self


class Memory(TableEngine):
    pass


class Log(TableEngine):
    pass


class StripeLog(TableEngine):
    pass


class TinyLog(TableEngine):
    pass


class Null(TableEngine):
    pass


class Set(TableEngine):
    pass


class Dictionary(TableEngine):
    arg_names = ['dictionary']

    # pylint: disable=unused-argument
    def __init__(self, dictionary: str = None):
        super().__init__(locals())


class Merge(TableEngine):
    arg_names = ['db_name, tables_regexp']

    # pylint: disable=unused-argument
    def __init__(self, db_name: str = None, tables_regexp: str = None):
        super().__init__(locals())


class File(TableEngine):
    arg_names = ['fmt']

    # pylint: disable=unused-argument
    def __init__(self, fmt: str = None):
        super().__init__(locals())


class Distributed(TableEngine):
    arg_names = ['cluster', 'database', 'table', 'sharding_key', 'policy_name']
    optional_args = {'sharding_key', 'policy_name'}

    # pylint: disable=unused-argument
    def __init__(self, cluster: str = None, database: str = None, table=None,
                 sharding_key: str = None, policy_name: str = None):
        super().__init__(locals())


class MergeTree(TableEngine):
    eng_params = ['order_by', 'partition_key', 'primary_key', 'sample_by']

    # pylint: disable=unused-argument
    def __init__(self, order_by: str = None, primary_key: str = None,
                 partition_by: str = None, sample_by: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


class SummingMergeTree(MergeTree):
    pass


class AggregatingMergeTree(MergeTree):
    pass


class ReplacingMergeTree(TableEngine):
    arg_names = ['ver']
    optional_args = set(arg_names)
    eng_params = MergeTree.eng_params

    # pylint: disable=unused-argument
    def __init__(self, ver: str = None, order_by: str = None, primary_key: str = None,
                 partition_by: str = None, sample_by: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


class CollapsingMergeTree(TableEngine):
    arg_names = ['sign']
    eng_params = MergeTree.eng_params

    # pylint: disable=unused-argument
    def __init__(self, sign: str = None, order_by: str = None, primary_key: str = None,
                 partition_by: str = None, sample_by: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


class VersionedCollapsingMergeTree(TableEngine):
    arg_names = ['sign', 'version']
    eng_params = MergeTree.eng_params

    # pylint: disable=unused-argument
    def __init__(self, sign: str = None, version: str = None,order_by: str = None, primary_key: str = None,
                 partition_by: str = None, sample_by: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


class GraphiteMergeTree(TableEngine):
    arg_names = ['config_section']
    eng_params = MergeTree.eng_params

    # pylint: disable=unused-argument
    def __init__(self, config_section: str = None, version: str = None,order_by: str = None, primary_key: str = None,
                 partition_by: str = None, sample_by: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


class ReplicatedMergeTree(TableEngine):
    arg_names = ['zk_path', 'replica']
    quoted_args = set(arg_names)
    optional_args = quoted_args
    eng_params = MergeTree.eng_params

    # pylint: disable=unused-argument
    def __init__(self, order_by: str = None, primary_key: str = None, partition_by: str = None, sample_by: str = None,
                 zk_path: str = None, replica: str = None):
        if not order_by and not primary_key:
            raise ArgumentError(None, 'Either PRIMARY KEY or ORDER BY must be specified')
        super().__init__(locals())


def build_engine(full_engine: str) -> Optional[TableEngine]:
    """
    Factory function to create TableEngine class from ClickHouse full_engine expression
    :param full_engine
    :return: TableEngine DDL element
    """
    if not full_engine:
        return None
    name = full_engine.split(' ')[0].split('(')[0]
    try:
        engine_cls = engine_map[name]
    except KeyError:
        if not name.startswith('System'):
            logger.warning('Engine %s not found', name)
        return None
    engine = engine_cls.__new__(engine_cls)
    engine.name = name
    engine.full_engine = full_engine
    return engine
